@@ -13,20 +13,30 @@ from typing import (
 from . import fast_data_types as defines
 from .conf.definition import Option, Shortcut, option_func
 from .conf.utils import (
-    choices, positive_float, positive_int, to_bool, to_cmdline as tc, to_color,
-    to_color_or_none, unit_float
+    choices, to_bool, to_cmdline as tc, to_color, to_color_or_none, unit_float
 )
-from .constants import FloatEdges, config_dir, is_macos
+from .constants import config_dir, is_macos
 from .fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE
+from .key_names import (
+    character_key_name_aliases, functional_key_name_aliases,
+    get_key_name_lookup
+)
 from .layout.interface import all_layouts
 from .rgb import Color, color_as_int, color_as_sharp, color_from_int
-from .utils import log_error
+from .types import FloatEdges, SingleKey
+from .utils import log_error, positive_float, positive_int
+
+
+class InvalidMods(ValueError):
+    pass
+
 
 MINIMUM_FONT_SIZE = 4
-
-
 mod_map = {'CTRL': 'CONTROL', 'CMD': 'SUPER', '⌘': 'SUPER',
            '⌥': 'ALT', 'OPTION': 'ALT', 'KITTY_MOD': 'KITTY'}
+character_key_name_aliases_with_ascii_lowercase = character_key_name_aliases.copy()
+for x in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
+    character_key_name_aliases_with_ascii_lowercase[x] = x.lower()
 
 
 def to_cmdline(x: str) -> List[str]:
@@ -43,7 +53,8 @@ def parse_mods(parts: Iterable[str], sc: str) -> Optional[int]:
         try:
             mods |= getattr(defines, 'GLFW_MOD_' + map_mod(m.upper()))
         except AttributeError:
-            log_error('Shortcut: {} has unknown modifier, ignoring'.format(sc))
+            if m.upper() != 'NONE':
+                log_error('Shortcut: {} has unknown modifier, ignoring'.format(sc))
             return None
 
     return mods
@@ -53,6 +64,42 @@ def to_modifiers(val: str) -> int:
     return parse_mods(val.split('+'), val) or 0
 
 
+def parse_shortcut(sc: str) -> SingleKey:
+    if sc.endswith('+') and len(sc) > 1:
+        sc = sc[:-1] + 'plus'
+    parts = sc.split('+')
+    mods = 0
+    if len(parts) > 1:
+        mods = parse_mods(parts[:-1], sc) or 0
+        if not mods:
+            raise InvalidMods('Invalid shortcut')
+    q = parts[-1]
+    q = character_key_name_aliases_with_ascii_lowercase.get(q.upper(), q)
+    is_native = False
+    if q.startswith('0x'):
+        try:
+            key = int(q, 16)
+        except Exception:
+            key = 0
+        else:
+            is_native = True
+    else:
+        try:
+            key = ord(q)
+        except Exception:
+            uq = q.upper()
+            uq = functional_key_name_aliases.get(uq, uq)
+            x: Optional[int] = getattr(defines, f'GLFW_FKEY_{uq}', None)
+            if x is None:
+                lf = get_key_name_lookup()
+                key = lf(q, False) or 0
+                is_native = key > 0
+            else:
+                key = x
+
+    return SingleKey(mods, is_native, key or 0)
+
+
 T = TypeVar('T')
 
 
@@ -86,8 +133,8 @@ o, k, g, all_groups = option_func(all_options, {
     'colors.table': [
         _('The color table'),
         _('''\
-The 16 terminal colors. There are 8 basic colors, each color has a dull and
-bright version. You can also set the remaining colors from the 256 color table
+The 256 terminal colors. There are 8 basic colors, each color has a dull and
+bright version, for the first 16 colors. You can set the remaining 240 colors
 as color16 to color255.''')
     ],
     'advanced': [_('Advanced')],
@@ -95,9 +142,9 @@ as color16 to color255.''')
     'shortcuts': [
         _('Keyboard shortcuts'),
         _('''\
-For a list of key names, see: :link:`the GLFW key macros
-<https://github.com/kovidgoyal/kitty/blob/master/glfw/glfw3.h#L349>`.
-The name to use is the part after the :code:`GLFW_KEY_` prefix.
+Keys are identified simply by their lowercase unicode characters. For example:
+``a`` for the A key, ``[`` for the left square bracket key, etc. For functional
+keys, such as ``Enter or Escape`` the names are present at :ref:`functional`.
 For a list of modifier names, see:
 :link:`GLFW mods <https://www.glfw.org/docs/latest/group__mods.html>`
 
@@ -325,6 +372,9 @@ Note that this code is indexed by PostScript name, and not the font
 family. This allows you to define very precise feature settings; e.g. you can
 disable a feature in the italic font but not in the regular font.
 
+On Linux, these are read from the FontConfig database first and then this,
+setting is applied, so they can be configured in a single, central place.
+
 To get the PostScript name for a font, use :code:`kitty + list-fonts --psnames`:
 
 .. code-block:: sh
@@ -465,6 +515,10 @@ The current implementation stores the data in UTF-8, so approximatively
 10000 lines per megabyte at 100 chars per line, for pure ASCII text, unformatted text.
 A value of zero or less disables this feature. The maximum allowed size is 4GB.'''))
 
+o('scrollback_fill_enlarged_window', False, long_text=_('''
+Fill new space with lines from the scrollback buffer after enlarging a window.
+'''))
+
 o('wheel_scroll_multiplier', 5.0, long_text=_('''
 Modify the amount scrolled by the mouse wheel. Note this is only used for low
 precision scrolling devices, not for high precision scrolling on platforms such
@@ -517,9 +571,14 @@ def url_prefixes(x: str) -> Tuple[str, ...]:
     return tuple(a.lower() for a in x.replace(',', ' ').split())
 
 
-o('url_prefixes', 'http https file ftp', option_type=url_prefixes, long_text=_('''
+o('url_prefixes', 'http https file ftp gemini irc gopher mailto news git', option_type=url_prefixes, long_text=_('''
 The set of URL prefixes to look for when detecting a URL under the mouse cursor.'''))
 
+o('detect_urls', True, long_text=_('''
+Detect URLs under the mouse. Detected URLs are highlighted
+with an underline and the mouse cursor becomes a hand over them.
+Even if this option is disabled, URLs are still clickable.'''))
+
 
 def copy_on_select(raw: str) -> str:
     q = raw.lower()
@@ -862,6 +921,8 @@ o('tab_bar_style', 'fade', option_type=choices('fade', 'separator', 'powerline',
 The tab bar style, can be one of: :code:`fade`, :code:`separator`, :code:`powerline`, or :code:`hidden`.
 In the fade style, each tab's edges fade into the background color, in the separator style, tabs are
 separated by a configurable separator, and the powerline shows the tabs as a continuous line.
+If you use the hidden style, you might want to create a mapping for the :code:`select_tab` action which
+presents you with a list of tabs and allows for easy switching to a tab.
 '''))
 
 
@@ -873,11 +934,12 @@ o('tab_bar_min_tabs', 2, option_type=tab_bar_min_tabs, long_text=_('''
 The minimum number of tabs that must exist before the tab bar is shown
 '''))
 
-o('tab_switch_strategy', 'previous', option_type=choices('previous', 'left', 'last'), long_text=_('''
+o('tab_switch_strategy', 'previous', option_type=choices('previous', 'left', 'right', 'last'), long_text=_('''
 The algorithm to use when switching to a tab when the current tab is closed.
 The default of :code:`previous` will switch to the last used tab. A value of
 :code:`left` will switch to the tab to the left of the closed tab. A value
-of :code:`last` will switch to the right-most tab.
+of :code:`right` will switch to the tab to the right of the closed tab.
+A value of :code:`last` will switch to the right-most tab.
 '''))
 
 
@@ -896,6 +958,11 @@ entries to this list.
 o('tab_separator', '"{}"'.format(default_tab_separator), option_type=tab_separator, long_text=_('''
 The separator between tabs in the tab bar when using :code:`separator` as the :opt:`tab_bar_style`.'''))
 
+o('tab_powerline_style', 'angled', option_type=choices('angled', 'slanted', 'round'), long_text=_('''
+The powerline separator style between tabs in the tab bar when using :code:`powerline`
+as the :opt:`tab_bar_style`, can be one of: :code:`angled`, :code:`slanted`, or :code:`round`.
+'''))
+
 
 def tab_activity_symbol(x: str) -> Optional[str]:
     if x == 'none':
@@ -932,6 +999,10 @@ layout name and :code:`{num_windows}` for the number of windows
 in the tab. Note that formatting is done by Python's string formatting
 machinery, so you can use, for instance, :code:`{layout_name[:2].upper()}` to
 show only the first two letters of the layout name, upper-cased.
+If you want to style the text, you can use styling directives, for example:
+:code:`{fmt.fg.red}red{fmt.fg.default}normal{fmt.bg._00FF00}green bg{fmt.bg.normal}`.
+Similarly, for bold and italic:
+:code:`{fmt.bold}bold{fmt.nobold}normal{fmt.italic}italic{fmt.noitalic}`.
 '''))
 o('active_tab_title_template', 'none', option_type=active_tab_title_template, long_text=_('''
 Template to use for active tabs, if not specified falls back
@@ -1010,14 +1081,7 @@ o('dim_opacity', 0.75, option_type=unit_float, long_text=_('''
 How much to dim text that has the DIM/FAINT attribute set. One means no dimming and
 zero means fully dimmed (i.e. invisible).'''))
 
-
-def selection_foreground(x: str) -> Optional[Color]:
-    if x.lower() != 'none':
-        return to_color(x)
-    return None
-
-
-o('selection_foreground', '#000000', option_type=selection_foreground, long_text=_('''
+o('selection_foreground', '#000000', option_type=to_color_or_none, long_text=_('''
 The foreground for text selected with the mouse. A value of none means to leave the color unchanged.'''))
 o('selection_background', '#fffacd', option_type=to_color, long_text=_('''
 The background for text selected with the mouse.'''))
@@ -1196,6 +1260,14 @@ def macos_titlebar_color(x: str) -> int:
     return (color_as_int(to_color(x)) << 8) | 2
 
 
+o('wayland_titlebar_color', 'system', option_type=macos_titlebar_color, long_text=_('''
+Change the color of the kitty window's titlebar on Wayland systems with client side window decorations such as GNOME.
+A value of :code:`system` means to use the default system color,
+a value of :code:`background` means to use the background color
+of the currently active window and finally you can use
+an arbitrary color, such as :code:`#12af59` or :code:`red`.
+'''))
+
 o('macos_titlebar_color', 'system', option_type=macos_titlebar_color, long_text=_('''
 Change the color of the kitty window's titlebar on macOS. A value of :code:`system`
 means to use the default system color, a value of :code:`background` means to use
@@ -1231,7 +1303,7 @@ both Option keys as Alt, instead.
 '''))
 
 o('macos_hide_from_tasks', False, long_text=_('''
-Hide the kitty window from running tasks (:kbd:`Option+Tab`) on macOS.
+Hide the kitty window from running tasks (:kbd:`⌘+Tab`) on macOS.
 '''))
 
 o('macos_quit_when_last_window_closed', False, long_text=_('''
@@ -1433,12 +1505,12 @@ if is_macos:
 g('shortcuts.tab')  # {{{
 k('next_tab', 'kitty_mod+right', 'next_tab', _('Next tab'))
 if is_macos:
-    k('next_tab', 'ctrl+tab', 'next_tab', _('Next tab'), add_to_docs=False)
     k('next_tab', 'shift+cmd+]', 'next_tab', _('Next tab'), add_to_docs=False)
+    k('next_tab', 'ctrl+tab', 'next_tab', _('Next tab'), add_to_docs=False)
 k('previous_tab', 'kitty_mod+left', 'previous_tab', _('Previous tab'))
 if is_macos:
-    k('previous_tab', 'shift+ctrl+tab', 'previous_tab', _('Previous tab'), add_to_docs=False)
     k('previous_tab', 'shift+cmd+[', 'previous_tab', _('Previous tab'), add_to_docs=False)
+    k('previous_tab', 'shift+ctrl+tab', 'previous_tab', _('Previous tab'), add_to_docs=False)
 k('new_tab', 'kitty_mod+t', 'new_tab', _('New tab'))
 if is_macos:
     k('new_tab', 'cmd+t', 'new_tab', _('New tab'), add_to_docs=False)
@@ -1459,12 +1531,17 @@ k('next_layout', 'kitty_mod+l', 'next_layout', _('Next layout'))
 
 g('shortcuts.fonts')  # {{{
 k('increase_font_size', 'kitty_mod+equal', 'change_font_size all +2.0', _('Increase font size'))
+k('increase_font_size', 'kitty_mod+plus', 'change_font_size all +2.0', _('Increase font size'), add_to_docs=False)
+k('increase_font_size', 'kitty_mod+kp_add', 'change_font_size all +2.0', _('Increase font size'), add_to_docs=False)
 if is_macos:
     k('increase_font_size', 'cmd+plus', 'change_font_size all +2.0', _('Increase font size'), add_to_docs=False)
+    k('increase_font_size', 'cmd+equal', 'change_font_size all +2.0', _('Increase font size'), add_to_docs=False)
     k('increase_font_size', 'cmd+shift+equal', 'change_font_size all +2.0', _('Increase font size'), add_to_docs=False)
 k('decrease_font_size', 'kitty_mod+minus', 'change_font_size all -2.0', _('Decrease font size'))
+k('decrease_font_size', 'kitty_mod+kp_subtract', 'change_font_size all -2.0', _('Decrease font size'))
 if is_macos:
     k('decrease_font_size', 'cmd+minus', 'change_font_size all -2.0', _('Decrease font size'), add_to_docs=False)
+    k('decrease_font_size', 'cmd+shift+minus', 'change_font_size all -2.0', _('Decrease font size'), add_to_docs=False)
 k('reset_font_size', 'kitty_mod+backspace', 'change_font_size all 0', _('Reset font size'))
 if is_macos:
     k('reset_font_size', 'cmd+0', 'change_font_size all 0', _('Reset font size'), add_to_docs=False)
@@ -1507,7 +1584,11 @@ g('shortcuts.misc')  # {{{
 k('toggle_fullscreen', 'kitty_mod+f11', 'toggle_fullscreen', _('Toggle fullscreen'))
 k('toggle_maximized', 'kitty_mod+f10', 'toggle_maximized', _('Toggle maximized'))
 k('input_unicode_character', 'kitty_mod+u', 'kitten unicode_input', _('Unicode input'))
+if is_macos:
+    k('input_unicode_character', 'cmd+ctrl+space', 'kitten unicode_input', _('Unicode input'), add_to_docs=False)
 k('edit_config_file', 'kitty_mod+f2', 'edit_config_file', _('Edit config file'))
+if is_macos:
+    k('edit_config_file', 'cmd+,', 'edit_config_file', _('Edit config file'), add_to_docs=False)
 k('kitty_shell', 'kitty_mod+escape', 'kitty_shell window', _('Open the kitty command shell'), long_text=_('''
 Open the kitty shell in a new window/tab/overlay/os_window to control kitty using commands.'''))
 k('increase_background_opacity', 'kitty_mod+a>m', 'set_background_opacity +0.1', _('Increase background opacity'))
